# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
#
# Modifications copyright (C) 2023 University of Illinois at Urbana-Champaign
# Author: Andrew Lytle


"""DenseGrouper Class"""

from typing import List, Tuple, Union, cast

from qiskit.opflow.converters.converter_base import ConverterBase
from qiskit.opflow.evolutions.evolved_op import EvolvedOp
from qiskit.opflow.exceptions import OpflowError
from qiskit.opflow.list_ops.list_op import ListOp
from qiskit.opflow.list_ops.summed_op import SummedOp
from qiskit.opflow.operator_base import OperatorBase
from qiskit.opflow.primitive_ops.pauli_op import PauliOp
from qiskit.opflow.primitive_ops.pauli_sum_op import PauliSumOp
from qiskit.opflow.state_fns.operator_state_fn import OperatorStateFn

from dense_ev.rmatrix import get_groups, get_groups2
from math import log

class DenseGrouper(ConverterBase):
    """The DenseGrouper converts SummedOps into a sum of sums, 
    organized by optimal dense family groupings.

    Meaning, it will traverse the Operator, and when it finds a SummedOp, it will evaluate which of
    the summed sub-Operators commute with one another. It will then convert each of the groups of
    commuting Operators into their own SummedOps, and return the sum-of-commuting-SummedOps.
    This is particularly useful for cases where mutually commuting groups can be handled
    similarly, as in the case of Pauli Expectations, where commuting Paulis have the same
    diagonalizing circuit rotation, or Pauli Evolutions, where commuting Paulis can be
    diagonalized together.
    """

    def __init__(self, traverse: bool = True) -> None:
        """
        Args:
            traverse: Whether to convert only the Operator passed to ``convert``, or traverse
                down that Operator.
        """
        self._traverse = traverse

    def convert(self, operator: OperatorBase) -> OperatorBase:
        """Check if operator is a SummedOp, in which case covert it into a sum of mutually
        commuting sums, or if the Operator contains sub-Operators and ``traverse`` is True,
        attempt to convert any sub-Operators.

        Args:
            operator: The Operator to attempt to convert.

        Returns:
            The converted Operator.
        """
        if isinstance(operator, PauliSumOp):
            return self.group_subops(operator)

        if isinstance(operator, ListOp):
            if isinstance(operator, SummedOp) and all(
                isinstance(op, PauliOp) for op in operator.oplist
            ):
                return self.group_subops(operator)
            elif self._traverse:
                return operator.traverse(self.convert)
        elif isinstance(operator, OperatorStateFn) and self._traverse:
            return OperatorStateFn(
                self.convert(operator.primitive),
                is_measurement=operator.is_measurement,
                coeff=operator.coeff,
            )
        elif isinstance(operator, EvolvedOp) and self._traverse:
            return EvolvedOp(self.convert(operator.primitive), coeff=operator.coeff)
        return operator

    @classmethod
    def group_subops(cls, list_op: Union[ListOp, PauliSumOp]) -> ListOp:
        """Given a ListOp, attempt to group into ListOps of the same type.

        Args:
            list_op: The Operator to group into dense groups

        Returns:
            The grouped Operator.

        Raises:
            OpflowError: If any of list_op's sub-ops is not ``PauliOp``.
        """
        if isinstance(list_op, ListOp):
            for op in list_op.oplist:
                if not isinstance(op, PauliOp):
                    raise OpflowError(
                        "Cannot determine groups if any Operator in list_op is not "
                        f"`PauliOp`. E.g., {op} ({type(op)})"
                    )

        if isinstance(list_op, PauliSumOp):
            # Roundabout way of getting m.
            primitive = list_op.primitive
            m = int(log(primitive[0].to_matrix().shape[0], 2))
            #print(f'{m = }')
            #print(f'{primitive = }')
            #print(f'{len(primitive) = }')
            if len(primitive) == 4**m:
                groups = get_groups2(list_op, m)
                result = SummedOp(
                    [PauliSumOp(primitive[group], grouping_type="TPB") for group in groups.values()],
                    coeff=list_op.coeff,
                )

            else:
                groups = get_groups2(list_op, m)
                result = SummedOp(
                    [PauliSumOp(primitive[group], grouping_type="TPB") for group in groups.values()],
                    coeff=list_op.coeff,
                )
            return result
        
        group_ops: List[ListOp] = [
            list_op.__class__([list_op[idx] for idx in group], abelian=True)
            for group in groups.values()
        ]
        if len(group_ops) == 1:
            return group_ops[0].mul(list_op.coeff)
        return list_op.__class__(group_ops, coeff=list_op.coeff)



#!/usr/bin/env python3
# Authors: Nouman Butt, Andrew Lytle

# Copyright 2023 University of Illinois at Urbana-Champaign
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import scipy.sparse as sp
import itertools, functools
#from e1plus import get_eigenvalues

op_I = sp.eye(2)
op_Z = sp.dia_matrix([[1,0],[0,-1]])
op_X = sp.dia_matrix([[0,1],[1,0]])
op_Y = -1j * op_Z @ op_X

C1 = [[1,0],[0,0]]
C2 = [[0,1],[0,0]]
C3 = [[0,0],[1,0]]
C4 = [[0,0],[0,1]]



C_ops = { "C1" : C1, "C2" :C2, "C3": C3, "C4" : C4 }
pauli_ops = { "I" : op_I, "Z" : op_Z, "X": op_X, "Y" : op_Y }

# def pauli_strings(num_qubits):
#     """Generate all Pauli strings with num_qubits.

#     Args: 
#         int, number of qubits
#     Returns:
#         Generator of Pauli strings, as tensor products.
#     """
#     for pauli_string in itertools.product(pauli_ops.keys(), repeat = num_qubits):
#         # construct this pauli string as a matrix
#         ops = [pauli_ops[tag] for tag in pauli_string]
#         op = functools.reduce(sp.kron, ops)
#         yield op.toarray()

def sp_kron_dok(mat_A, mat_B): 
    """Kronecker (tensor) product of two sparse matrices.
    
    Args:
        mat_A, mat_B: 2d numpy arrays
    Returns:
        Sparse matrix in "dictionary of keys" format (to eliminate zeros)
    """
    return sp.kron(mat_A, mat_B, format = "dok")

def to_pauli_vec(mat):
    """Pauli string decomposition of a matrix.

    Args:
        2d numpy array, mat
    Returns:
        Dictionary pauli_vec, e.g. {'XX': 0.5, 'XY': 0.5}
    """
    pauli_vec = {} # the dictionary we are saving

    mat_vec = np.array(mat).ravel()
    num_qubits = int(np.log2(np.sqrt(mat_vec.size)))

    for pauli_string in itertools.product(pauli_ops.keys(), repeat = num_qubits):
        # construct this pauli string as a matrix
        ops = [ pauli_ops[tag] for tag in pauli_string ]
        op = functools.reduce(sp_kron_dok, ops)

        # compute an inner product, same as tr(A @ B) but faster
        op_vec = op.transpose().reshape((1, 4**num_qubits))
        coefficient = (op_vec * mat_vec).sum() / 2**num_qubits
        pauli_vec["".join(pauli_string)] = coefficient

    return pauli_vec

def pauli_string_to_mat(pauli_string):
    """Numpy array representing pauli_string.
    
    Args: pauli_string e.g. 'XYZ'
    Returns: 
        Numpy array of shape (2**m, 2**m) representing pauli_string as 
        a tensor product.
    """
    ops_primitive = [pauli_ops[tag] for tag in pauli_string]
    op = functools.reduce(sp.kron, ops_primitive)
    return op.toarray()

# See also gen_random_pauli_dict(m) in Psfam_documentation.ipynb

def from_pauli_vec(pauli_vec):
    """Numpy array/matrix from dictionary of Pauli strings.

    Args:
        pauli_vec: Dictionary with Pauli string decomposition, 
        e.g. {'YZ': 1.0, 'XY': 2.0}.
    Returns:
        Numpy array corresponding to matrix specified by pauli_vec.
    """
    m = len(list(pauli_vec.keys())[0])
    result = np.zeros((2**m, 2**m), dtype='complex128')
    for string, coeff in pauli_vec.items():
        result += coeff*pauli_string_to_mat(string)
    return result

def to_C_vec(mat):
    C_vec = {} # the dictionary we are saving

    mat_vec = np.array(mat).ravel()
    num_qubits = int(np.log2(np.sqrt(mat_vec.size)))

    for C_string in itertools.product(C_ops.keys(), repeat = num_qubits):
        # construct this pauli string as a matrix
        ops = [ C_ops[tag] for tag in C_string ]
        op = functools.reduce(sp_kron_dok, ops)

        # compute an inner product, same as tr(A @ B) but faster
        op_vec = op.reshape((1,4**num_qubits))
        coefficient = ( op_vec * mat_vec ).sum() / 2**num_qubits
        if coefficient != 0:
            C_vec["".join(C_string)] = coefficient

    return C_vec


def test():
    mat = [[1,0,0,0],[0,0,0,1],[0,1,0,0],[0,1,0,0]] # for example...
    print(to_pauli_vec(mat))
    print(sp_kron_dok(mat, mat))

    #for ps in pauli_strings(1):
    #    print(ps)

    print()
    #pauli_vec = {'XY': 1.0, 'ZZ': 1.0}
    pauli_vec = {'Y': 1.0}
    matrix = from_pauli_vec(pauli_vec)
    print(matrix)
    print(to_pauli_vec(matrix))

    print(pauli_string_to_mat('Y'))
    print(op_Y.toarray())

if __name__ == "__main__":
    test()



#!/usr/bin/env python3
# Author: Andrew Lytle
# Dec 2022

# Copyright 2023 University of Illinois at Urbana-Champaign
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
stream_handler = logging.StreamHandler()
log_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
formatter = logging.Formatter(log_format)
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)

import numpy as np

from qiskit import Aer
from qiskit.opflow import StateFn, DictStateFn, PauliExpectation, CircuitSampler
from qiskit.quantum_info import random_statevector
from qiskit.utils import QuantumInstance

from dense_ev.rmatrix import random_H, get_Op
from dense_ev.dense_pauli_expectation import DensePauliExpectation
    
def main(m, optype, name, nshots):
    seed = 42
    np.random.seed(seed)
    N = 2**m
    Hmat = random_H(N)

    logger.debug(Hmat)

    H = get_Op(Hmat, 'naive')

    sf = StateFn(H)
    sf = sf.adjoint()

    state = DictStateFn(random_statevector(2**m).to_dict())
    _eval = sf.eval(state)
    logger.info(f"Direct evaluation: {_eval}")

    logger.debug("Building PauliExpectation")
    if optype == 'naive':
        expectation = PauliExpectation(group_paulis=False).convert(sf.compose(state))
    if optype == 'abelian':
        expectation = PauliExpectation(group_paulis=True).convert(sf.compose(state))
    if optype == 'dense':
        expectation = DensePauliExpectation(group_paulis=True).convert(sf.compose(state))
    
    logger.debug(f"Using backend {name}")
    backend = get_backend(name)
    qi = QuantumInstance(backend, shots=nshots, optimization_level=3)
    sampler = CircuitSampler(qi, attach_results=True).convert(expectation)

    logger.info(f"Direct evaluation: {_eval}")
    logger.info('RESULT:', sampler.eval())

def unit_test(m):
    N = 2**m
    Hmat = random_H(N)
    H = get_Op(Hmat, 'naive')
    sf = StateFn(H)
    sf = sf.adjoint()
    state = DictStateFn(random_statevector(2**m).to_dict())
    direct_eval = sf.eval(state)
    
    expectation = DensePauliExpectation(group_paulis=True).convert(sf.compose(state))
    backend = Aer.get_backend('statevector_simulator')
    qi = QuantumInstance(backend)
    sampler = CircuitSampler(qi, attach_results=True).convert(expectation)
    new_eval = sampler.eval()
    print(f"{direct_eval = }")
    print(f"{new_eval = }")
    return np.isclose(new_eval, direct_eval.real)

def run_unit_tests():
    results = []
    for m in range(1,4):
        for test in range(1, 11):
            print(f"{m = } , {test = }")
            passed = unit_test(m)
            results.append(passed)
            if passed:
                continue
            else:
                print(f"Test failed with {m = }")
                break
        if not passed:
            break
    print(f"Performed {len(results)} tests, "
        f"all tests passed = {np.array(results).all()==True}")
   
if __name__ == "__main__":
    run_unit_tests()
    



#!/usr/bin/env python3
# Author: Andrew Lytle
# Nov 2022

# Copyright 2023 University of Illinois at Urbana-Champaign
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import log
from collections import defaultdict

import numpy as np
from numpy.linalg import eig
from numpy.random import normal
from scipy.stats import unitary_group

from qiskit.opflow.primitive_ops import PauliOp
from qiskit.quantum_info.operators import Pauli
from qiskit.opflow.list_ops.summed_op import SummedOp
from qiskit.opflow.primitive_ops.pauli_sum_op import PauliSumOp
from qiskit.opflow.converters import AbelianGrouper

#from qiskit.opflow.converters.pauli_basis_change import PauliBasisChange
#from qiskit.opflow.state_fns.state_fn import StateFn

from dense_ev.decompose_pauli import to_pauli_vec
from psfam.pauli_organizer import PauliOrganizer

def random_evals(N):
    "List of Gaussian distributed numbers [(0,1),...]."
    vals = []
    for i in range(N):
        vals.append(normal())
    return vals

def dot_all(Ms):
    "Dot product of [M1, M2, ...]"
    res = np.identity(Ms[0].shape[0])
    for M in Ms[::-1]:
        res = np.dot(M, res)
    return res

def hc(M):
    "Hermitian conjugate of M."
    return M.conj().T

def random_H(N):
    "Random NxN Hermitian matrix."
    evs = random_evals(N)
    D = np.diag(evs)
    U = unitary_group.rvs(N)
    H = dot_all([U, D, hc(U)])
    return H

def test_random_H(m):
    N=pow(2,m)
    evs = random_evals(N)
    D = np.diag(evs)
    U = unitary_group.rvs(N)
    H = dot_all([U, D, hc(U)])
    vals, vecs = eig(H)

    print(evs)
    print(D)
    print(vals)

def array_to_Op(Hmat):
    "Convert numpy matrix to qiskit Operator type object."

    N = Hmat.shape[0]
    m = log(N, 2)
    assert m == int(m)
    m = int(m)

    pauli_vec = to_pauli_vec(Hmat)
    #print(pauli_vec)
    #print(len(pauli_vec))
    
    H_op=PauliOp(Pauli('I'*m), 0.0)
    #print(type(H_op))
    for pauli_string in pauli_vec.keys():
        coefficient = pauli_vec[pauli_string]
        #if(abs(coefficient) > 0.0001 ):
        H_op += PauliOp(Pauli(pauli_string), coefficient)
    #print(type(H_op))
    return H_op

def get_groups(m):
    """Specification of Pauli string families suitable for use in Qiskit.
    
    Args: 
        int m: Number of qubits
    
    Returns:
        defaultdict<list> {1: [i, j, k,...], 2: [l, m, n,...], ...}
        where [i, j, k,...] are integers specifying the Pauli operators
        in the family, according to the internal Qiskit ordering
        (e.g. id_list below.)

    Note: Currently function generates a random H matrix to obtain 
    the default qiskit operator ordering (H.primitive). This
    can probably be rewritten so that this isn't necessary.
    """
    N = 2**m
    PO = PauliOrganizer(m)

    # Long way to get primitive elements..
    Hmat = random_H(N)
    H = array_to_Op(Hmat)
    primitive = H.primitive
    #print('primitive:', primitive)
    #print('paulis:', primitive.paulis)
    
    # How stored in Op objects.
    '''
    id_list = \
    ['III', 'IIZ', 'IIX', 'IIY', 'IZI', 'IZZ', 'IZX', 'IZY', 'IXI', 'IXZ', 'IXX', 'IXY', 
    'IYI', 'IYZ', 'IYX', 'IYY', 'ZII', 'ZIZ', 'ZIX', 'ZIY', 'ZZI', 'ZZZ', 'ZZX', 'ZZY', 
    'ZXI', 'ZXZ', 'ZXX', 'ZXY', 'ZYI', 'ZYZ', 'ZYX', 'ZYY', 'XII', 'XIZ', 'XIX', 'XIY', 
    'XZI', 'XZZ', 'XZX', 'XZY', 'XXI', 'XXZ', 'XXX', 'XXY', 'XYI', 'XYZ', 'XYX', 'XYY', 
    'YII', 'YIZ', 'YIX', 'YIY', 'YZI', 'YZZ', 'YZX', 'YZY', 'YXI', 'YXZ', 'YXX', 'YXY', 
    'YYI', 'YYZ', 'YYX', 'YYY']
    '''
    # Will primitive.paulis include the full set irrespective of H?
    id_list = [str(x) for x in primitive.paulis]
    id_dict = { id_list[x]: x for x in range(len(id_list))}
    #print('id_dict:', id_dict)

    res = []
    for family in PO.f:
        #print(family.to_string())
        fam_ids = []
        for op in family.to_string():
            fam_ids.append(id_dict[op])
        res.append(fam_ids)
    # Feb 25, the following line isn't needed anymore?
    #res[-1].append(0)  # Add the identity operator to the last family.

    groups = defaultdict(list)
    groups = {i: res[i] for i in range(len(res))}
    return groups

def get_groups2(H, m):
    """Specification of Pauli string families suitable for use in Qiskit.
    
    Args: 
        int m: Number of qubits
    
    Returns:
        defaultdict<list> {1: [i, j, k,...], 2: [l, m, n,...], ...}
        where [i, j, k,...] are integers specifying the Pauli operators
        in the family, according to the internal Qiskit ordering
        (e.g. id_list below.)

    Note: Currently function generates a random H matrix to obtain 
    the default qiskit operator ordering (H.primitive). This
    can probably be rewritten so that this isn't necessary.
    """
    N = 2**m
    PO = PauliOrganizer(m)

    # Long way to get primitive elements..
    #Hmat = random_H(N)
    #H = array_to_Op_filter(Hmat)
    primitive = H.primitive
    #print('primitive:', primitive)
    #print('paulis:', primitive.paulis)
    
    # How stored in Op objects.
    '''
    id_list = \
    ['III', 'IIZ', 'IIX', 'IIY', 'IZI', 'IZZ', 'IZX', 'IZY', 'IXI', 'IXZ', 'IXX', 'IXY', 
    'IYI', 'IYZ', 'IYX', 'IYY', 'ZII', 'ZIZ', 'ZIX', 'ZIY', 'ZZI', 'ZZZ', 'ZZX', 'ZZY', 
    'ZXI', 'ZXZ', 'ZXX', 'ZXY', 'ZYI', 'ZYZ', 'ZYX', 'ZYY', 'XII', 'XIZ', 'XIX', 'XIY', 
    'XZI', 'XZZ', 'XZX', 'XZY', 'XXI', 'XXZ', 'XXX', 'XXY', 'XYI', 'XYZ', 'XYX', 'XYY', 
    'YII', 'YIZ', 'YIX', 'YIY', 'YZI', 'YZZ', 'YZX', 'YZY', 'YXI', 'YXZ', 'YXX', 'YXY', 
    'YYI', 'YYZ', 'YYX', 'YYY']
    '''
    # Will primitive.paulis include the full set irrespective of H?
    id_list = [str(x) for x in primitive.paulis]
    id_dict = { id_list[x]: x for x in range(len(id_list))}
    #print('id_dict:', id_dict)

    res = []
    for family in PO.f:
        #print(family.to_string())
        fam_ids = []
        for op in family.to_string():
            try:
                fam_ids.append(id_dict[op])
            except KeyError:
                continue
        res.append(fam_ids)
    # Feb 25, the following line isn't needed anymore?
    #res[-1].append(0)  # Add the identity operator to the last family.

    groups = defaultdict(list)
    groups = {i: res[i] for i in range(len(res)) if res[i]}
    return groups

def array_to_SummedOp(Hmat):
    "Convert numpy matrix to SummedOp grouped into Pauli-string families."
    N = Hmat.shape[0]
    m = log(N, 2)
    assert m == int(m)
    m = int(m)
    #PO = PauliOrganizer(m)

    H = array_to_Op(Hmat)
    primitive = H.primitive
    #print('primitive:', primitive)
    #print('paulis:', primitive.paulis)
    
    groups = get_groups(m)

    result = SummedOp(
        [PauliSumOp(primitive[group], grouping_type="TPB") for group in groups.values()],
        coeff=1)

    #print('result:', result)
    #print('len(result):', len(result))

    return result

def get_Op(Hmat, optype):
    "Qiskit operators from matrix array."
    if optype == 'naive':
        return array_to_Op(Hmat)
    elif optype == 'abelian':
        grouper = AbelianGrouper()
        H = array_to_Op(Hmat)
        return grouper.convert(H)
    elif optype == 'new':
        return array_to_SummedOp(Hmat)
    else:
        msg = f'optype {optype} not recognized [naive, abelian, new].'
        raise ValueError(msg)

if __name__ == '__main__':
    #print(get_groups(2))
    Hmat = random_H(8)
    #print(array_to_SummedOp(Hmat))
    print(get_Op(Hmat, 'abelian'))
    
